/*M///////////////////////////////////////////////////////////////////////////////////////
 //
 //  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
 //
 //  By downloading, copying, installing or using the software you agree to this license.
 //  If you do not agree to this license, do not download, install,
 //  copy or use the software.
 //
 //
 //                           License Agreement
 //                For Open Source Computer Vision Library
 //
 // Copyright (C) 2013, OpenCV Foundation, all rights reserved.
 // Third party copyrights are property of their respective owners.
 //
 // Redistribution and use in source and binary forms, with or without modification,
 // are permitted provided that the following conditions are met:
 //
 //   * Redistribution's of source code must retain the above copyright notice,
 //     this list of conditions and the following disclaimer.
 //
 //   * Redistribution's in binary form must reproduce the above copyright notice,
 //     this list of conditions and the following disclaimer in the documentation
 //     and/or other materials provided with the distribution.
 //
 //   * The name of the copyright holders may not be used to endorse or promote products
 //     derived from this software without specific prior written permission.
 //
 // This software is provided by the copyright holders and contributors "as is" and
 // any express or implied warranties, including, but not limited to, the implied
 // warranties of merchantability and fitness for a particular purpose are disclaimed.
 // In no event shall the Intel Corporation or contributors be liable for any direct,
 // indirect, incidental, special, exemplary, or consequential damages
 // (including, but not limited to, procurement of substitute goods or services;
 // loss of use, data, or profits; or business interruption) however caused
 // and on any theory of liability, whether in contract, strict liability,
 // or tort (including negligence or otherwise) arising in any way out of
 // the use of this software, even if advised of the possibility of such damage.
 //
 //M*/
#define _USE_MATH_DEFINES

#include "precomp.hpp"
#include <math.h>

using namespace cv;

/* return default parameters */
BinaryDescriptor::Params::Params()
{
    LowestThreshold = 0.35;
    NNDRThreshold = 0.6;
    ksize_ = 5;
    numOfOctave_ = 5;
    numOfBand_ = 9;
    widthOfBand_ = 7;
    reductionRatio = sqrt(2); // or 2?

}

/* read parameters from a FileNode object and store them (struct function) */
void BinaryDescriptor::Params::read(const cv::FileNode& fn )
{
    LowestThreshold = fn["LowestThreshold"];
    NNDRThreshold = fn["NNDRThreshold"];
    ksize_ = fn["ksize_"];
    numOfOctave_ = fn["numOfOctave_"];
    numOfBand_ = fn["numOfBand_"];
    widthOfBand_ = fn["widthOfBand_"];
    reductionRatio = fn["reductionRatio"];
}

/* store parameters to a FileStorage object (struct function) */
void BinaryDescriptor::Params::write(cv::FileStorage& fs) const
{
    fs << "LowestThreshold" << LowestThreshold;
    fs << "NNDRThreshold" << NNDRThreshold;
    fs << "ksize_" <<  ksize_;
    fs << "numOfOctave_" << numOfOctave_;
    fs << "numOfBand_" <<  numOfBand_;
    fs << "widthOfBand_" << widthOfBand_;
    fs << "reductionRatio" << reductionRatio;
}

/* construct a BinaryDescrptor object and compute external private parameters */
BinaryDescriptor::BinaryDescriptor(const BinaryDescriptor::Params &parameters) : params(parameters)
{

    /* prepare a vector to host local weights F_l*/
    gaussCoefL_.resize(params.widthOfBand_*3);

    /* compute center of central band (every computation involves 2-3 bands) */
    double u = (params.widthOfBand_*3-1)/2;

    /* compute exponential part of F_l */
    double sigma = (params.widthOfBand_*2+1)/2;// (widthOfBand_*2+1)/2;
    double invsigma2 = -1/(2*sigma*sigma);

    /* compute all local weights */
    double dis;
    for(int i=0; i<params.widthOfBand_*3; i++)
    {
      dis = i-u;
      gaussCoefL_[i] = exp(dis*dis*invsigma2);
    }

    /* prepare a vector for global weights F_g*/
    gaussCoefG_.resize(params.numOfBand_*params.widthOfBand_);

    /* compute center of LSR */
    u = (params.numOfBand_*params.widthOfBand_-1)/2;

    /* compute exponential part of F_g */
    sigma = u;
    invsigma2 = -1/(2*sigma*sigma);
    for(int i=0; i<params.numOfBand_*params.widthOfBand_; i++)
    {
     dis = i-u;
     gaussCoefG_[i] = exp(dis*dis*invsigma2);
    }
}

/* read parameters from a FileNode object and store them (class function ) */
void BinaryDescriptor::read( const cv::FileNode& fn )
{
    params.read(fn);
}

/* store parameters to a FileStorage object (class function) */
void BinaryDescriptor::write( cv::FileStorage& fs ) const
{
    params.write(fs);
}

/* get coefficients of line passing by two points in line_extremes */
void BinaryDescriptor::getLineParameters(cv::Vec4i &line_extremes, cv::Vec3i &lineParams)
{
    int x1 = line_extremes[0];
    int x2 = line_extremes[2];
    int y1 = line_extremes[1];
    int y2 = line_extremes[3];

    /* line is parallel to Y axis */
    if(x1 == x2)
    {
        lineParams[0] = 1;
        lineParams[1] = 0;
        lineParams[2] = -x1 /* or -x2 */;
    }

    /* line is parallel to X axis */
    else if(y1 == y2)
    {
        lineParams[0] = 0;
        lineParams[1] = 1;
        lineParams[2] = -y1 /* or -y2 */;
    }

    /* line is not parallel to any axis */
    else
    {
        lineParams[0] = y1-y2;
        lineParams[1] = x2-x1;
        lineParams[2] = -y1*(x2-x1) + x1*(y2-y1);
    }
}
 /* compute the angle between line and X axis */
float BinaryDescriptor::getLineDirection(cv::Vec3i &lineParams)
{
    /* line is parallel to X axis */
    if(lineParams[0] == 0)
        return 0;

    /* line is parallel to Y axis */
    else if(lineParams[1] == 0)
        return M_PI/2;

    /* line is not parallel to any axis */
    else
    {
        /* compute angular coefficient */
        float m = -lineParams[0]/lineParams[1];

        if(m>0)
            return atan(m);

        else
            return M_PI - atan(m);
    }
}

/* extract lines from an image and compute their descriptors */
void BinaryDescriptor::getLineBinaryDescriptors(cv::Mat &oct_binaryDescMat)
{
    /*  start function that actually implements descriptors' computation */
    getLineBinaryDescriptorsImpl(oct_binaryDescMat);

}

/* compute descriptors */
void BinaryDescriptor::getLineBinaryDescriptorsImpl(cv::Mat &oct_binaryDescMat)
{
    /* prepare a matrix to store Gaussian pyramid of input matrix */
    std::vector<cv::Mat> matVec(params.numOfOctave_);

    /* reinitialize structures for hosting images' derivatives and sizes
    (they may have been used in the past) */
    dxImg_vector.clear();
    dyImg_vector.clear();
    images_sizes.clear();

    dxImg_vector.resize(params.numOfOctave_);
    dyImg_vector.resize(params.numOfOctave_);
    images_sizes.resize(params.numOfOctave_);

    /* insert input image into pyramid */
    cv::Mat currentMat = oct_binaryDescMat.clone();
    matVec.push_back(currentMat);
    images_sizes.push_back(currentMat.size());

    /* compute and store derivatives of input image */
    cv:Mat currentDx, currentDy;
    cv::Sobel( currentMat, currentDx, CV_16SC1, 1, 0, 3);
    cv::Sobel( currentMat, currentDy, CV_16SC1, 0, 1, 3);

    dxImg_vector.push_back(currentDx);
    dyImg_vector.push_back(currentDy);

    /* fill Gaussian pyramid */
    for(int i = 1; i<params.numOfOctave_; i++)
    {
        /* compute and store next image in pyramid and its size */
        pyrDown( currentMat, currentMat, Size( currentMat.cols/params.reductionRatio, currentMat.rows/params.reductionRatio ));
        matVec.push_back(currentMat);
        images_sizes.push_back(currentMat.size());

        /* compute and store derivatives of new image */
        cv::Sobel( currentMat, currentDx, CV_16SC1, 1, 0, 3);
        cv::Sobel( currentMat, currentDy, CV_16SC1, 0, 1, 3);

        dxImg_vector.push_back(currentDx);
        dyImg_vector.push_back(currentDy);
    }

    /* prepare a structure for hosting and organizing extracted lines */
    ScaleLines keyLines;

    /* extract and arrange lines */
    OctaveKeyLines(matVec, keyLines);

    /* compute LBD descriptors */
    ComputeLBD_(keyLines);

}

/* power function with error management */
static inline int get2Pow(int i) {
    if(i>=0 && i<=7)
        return pow(2,i);

    else
    {
        CV_Assert(false);
        return -1;
    }
}

/* conversion of an LBD descriptor to the decimal equivalent of its binary representation */
unsigned char BinaryDescriptor::binaryTest(float* f1, float* f2)
{
    uchar result = 0;
     for(int i = 0; i<8; i++)
     {
         if(f1[i]>f2[i])
             result+=get2Pow(i);
     }

     return result;

}

/* gather lines in groups. Each group contains the same line, detected in different octaves */
int BinaryDescriptor::OctaveKeyLines(std::vector<cv::Mat> & octaveImages, ScaleLines &keyLines)
{

    /* final number of extracted lines */
    unsigned int numOfFinalLine = 0;

    for(size_t scaleCounter = 0; scaleCounter<octaveImages.size(); scaleCounter++)
    {
        /* get current scaled image */
        cv::Mat currentScaledImage = octaveImages[scaleCounter];

        /* create an LSD detector and store a pointer to it */
        cv::Ptr<cv::LineSegmentDetector> ls = cv::createLineSegmentDetector(cv::LSD_REFINE_STD);

        /* prepare a vectore to host extracted segments */
        std::vector<cv::Vec4i> lines_std;

        /* use detector to extract segments */
        ls->detect(currentScaledImage, lines_std);

        /* store lines extracted from current image */
        extractedLines.push_back(lines_std);

        /* update lines counter */
        numOfFinalLine += lines_std.size();

    }

    /* prepare a vector to store octave information associated to extracted lines */
    std::vector<OctaveLine> octaveLines(numOfFinalLine);

    /* set lines' counter to 0 for reuse */
    numOfFinalLine = 0;

    /* counter to give a unique ID to lines in LineVecs */
    unsigned int lineIDInScaleLineVec = 0;

    /* floats to compute lines' lengths */
    float dx, dy;

    /* loop over lines extracted from scale 0 (original image) */
    for(unsigned int lineCurId=0; lineCurId<extractedLines[0].size(); lineCurId++)
    {
        /* set octave from which it was extracted */
        octaveLines[numOfFinalLine].octaveCount = 0;
        /* set ID within its octave */
        octaveLines[numOfFinalLine].lineIDInOctave = lineCurId;
        /* set a unique ID among all lines extracted in all octaves */
        octaveLines[numOfFinalLine].lineIDInScaleLineVec = lineIDInScaleLineVec;

        /* compute absolute value of difference between X coordinates of line's extreme points */
        dx = (extractedLines[0][lineCurId])[0] - (extractedLines[0][lineCurId])[2];
        /* compute absolute value of difference between Y coordinates of line's extreme points */
        dy = (extractedLines[0][lineCurId])[1] - (extractedLines[0][lineCurId])[3];
        /* compute line's length */
        octaveLines[numOfFinalLine].lineLength = sqrt(dx*dx+dy*dy);

        /* update counters */
        numOfFinalLine++;
        lineIDInScaleLineVec++;
    }

    /* create and fill an array to store scale factors */
    float *scale = new float[params.numOfOctave_];
    scale[0] = 1;
    for(unsigned int octaveCount = 1; octaveCount<params.numOfOctave_; octaveCount++ )
    {
            scale[octaveCount] = params.reductionRatio * scale[octaveCount-1];
    }

    /* some variables' declarations */
    float rho1, rho2, tempValue;
    float direction, near, length;
    unsigned int octaveID, lineIDInOctave;

    /*more than one octave image, organize lines in scale space.
    *lines corresponding to the same line in octave images should have the same index in the ScaleLineVec */
    if(params.numOfOctave_>1)
    {
        /* some other variables' declarations */
        float twoPI = 2*M_PI;
        unsigned int closeLineID;
        float endPointDis,minEndPointDis,minLocalDis,maxLocalDis;
        float lp0,lp1, lp2, lp3, np0,np1, np2, np3;

        /* loop over list of octaves */
                for(unsigned int octaveCount = 1; octaveCount<params.numOfOctave_; octaveCount++)
                {
                    /*for each line in current octave image, find their corresponding lines
                    in the octaveLines,
                    give them the same value of lineIDInScaleLineVec*/

                    /* loop over list of lines extracted from current octave */
                    for(unsigned int lineCurId=0;lineCurId<extractedLines[octaveCount].size();lineCurId++)
                    {
                        /* get (scaled) known term from equation of current line */
                        cv::Vec3i line_equation;
                        getLineParameters(extractedLines[octaveCount][lineCurId], line_equation);
                        rho1 = scale[octaveCount] *  fabs(line_equation[2]);

                        /*nearThreshold depends on the distance of the image coordinate origin to current line.
                        *so nearThreshold = rho1 * nearThresholdRatio, where nearThresholdRatio = 1-cos(10*pi/180) = 0.0152*/
                        tempValue = rho1 * 0.0152;
                        float nearThreshold = (tempValue>6)?(tempValue):6;
                        nearThreshold = (nearThreshold<12)?nearThreshold:12;

                        /* compute scaled lenght of current line */
                        dx = fabs((extractedLines[octaveCount][lineCurId])[0]-(extractedLines[octaveCount][lineCurId][2]));//x1-x2
                        dy = fabs((extractedLines[octaveCount][lineCurId])[1]-(extractedLines[octaveCount][lineCurId][3]));//y1-y2
                        length = scale[octaveCount] * sqrt(dx*dx+dy*dy);

                        minEndPointDis = 12;
                        /* loop over the octave representations of all lines */
                        for(unsigned int lineNextId=0; lineNextId<numOfFinalLine;lineNextId++)
                        {
                            /* if a line from same octave is encountered,
                            a comparison with it shouldn't be considered */
                            octaveID = octaveLines[lineNextId].octaveCount;
                            if(octaveID==octaveCount)
                                break;

                            /* take ID (in octave) of line to be compared */
                            lineIDInOctave = octaveLines[lineNextId].lineIDInOctave;

                            /* compute difference between lines' direction, to check
                            whether they are parallel */
                            cv::Vec3i line_equation_to_compare;
                            getLineParameters(extractedLines[octaveID][lineIDInOctave],
                                              line_equation_to_compare);

                            direction = fabs(getLineDirection(line_equation) -
                                             getLineDirection(line_equation_to_compare));

                            /* the angle between two lines are larger than 10degrees
                            (i.e. 10*pi/180=0.1745), they are not close to parallel */
                            if(direction>0.1745 && (twoPI - direction>0.1745))
                                continue;

                            /*now check whether current line and next line are near to each other.
                            Get known term from equation to be compared */
                            rho2 = scale[octaveID] * fabs(line_equation_to_compare[2]);

                            /* compute difference between known terms */
                            near = fabs(rho1 - rho2);

                            /* two lines are not close in the image */
                            if(near>nearThreshold)
                                continue;

                            /* get the extreme points of the two lines */
                            lp0 = scale[octaveCount] * (extractedLines[octaveCount][lineCurId])[0];
                            lp1 = scale[octaveCount] * (extractedLines[octaveCount][lineCurId])[1];
                            lp2 = scale[octaveCount] * (extractedLines[octaveCount][lineCurId])[2];
                            lp3 = scale[octaveCount] * (extractedLines[octaveCount][lineCurId])[3];
                            np0 = scale[octaveID] * (extractedLines[octaveID][lineIDInOctave])[0];
                            np1 = scale[octaveID] * (extractedLines[octaveID][lineIDInOctave])[1];
                            np2 = scale[octaveID] * (extractedLines[octaveID][lineIDInOctave])[2];
                            np3 = scale[octaveID] * (extractedLines[octaveID][lineIDInOctave])[3];

                            /* get the distance between the two leftmost extremes of lines
                            L1(0,1)<->L2(0,1) */
                            dx = lp0 - np0;
                            dy = lp1 - np1;
                            endPointDis = sqrt(dx*dx + dy*dy);

                            /* set momentaneously min and max distance between lines to
                            the one between left extremes */
                            minLocalDis = endPointDis;
                            maxLocalDis = endPointDis;

                            /* compute distance between right extremes
                            L1(2,3)<->L2(2,3) */
                            dx = lp2 - np2;
                            dy = lp3 - np3;
                            endPointDis = sqrt(dx*dx + dy*dy);

                            /* update (if necessary) min and max distance between lines */
                            minLocalDis = (endPointDis<minLocalDis)?endPointDis:minLocalDis;
                            maxLocalDis = (endPointDis>maxLocalDis)?endPointDis:maxLocalDis;

                            /* compute distance between left extreme of current line and
                            right extreme of line to be compared
                            L1(0,1)<->L2(2,3) */
                            dx = lp0 - np2;
                            dy = lp1 - np3;
                            endPointDis = sqrt(dx*dx + dy*dy);

                            /* update (if necessary) min and max distance between lines */
                            minLocalDis = (endPointDis<minLocalDis)?endPointDis:minLocalDis;
                            maxLocalDis = (endPointDis>maxLocalDis)?endPointDis:maxLocalDis;

                            /* compute distance between right extreme of current line and
                            left extreme of line to be compared
                            L1(2,3)<->L2(0,1) */
                            dx = lp2 - np0;
                            dy = lp3 - np1;
                            endPointDis = sqrt(dx*dx + dy*dy);

                            /* update (if necessary) min and max distance between lines */
                            minLocalDis = (endPointDis<minLocalDis)?endPointDis:minLocalDis;
                            maxLocalDis = (endPointDis>maxLocalDis)?endPointDis:maxLocalDis;

                            /* check whether conditions for considering line to be compared
                            worth to be inserted in the same LineVec are satisfied */
                            if((maxLocalDis<0.8*(length+octaveLines[lineNextId].lineLength))
                                &&(minLocalDis<minEndPointDis))
                            {
                                /* keep the closest line */
                                minEndPointDis = minLocalDis;
                                closeLineID = lineNextId;
                            }


                        }

                        /* add current line into octaveLines */
                        if(minEndPointDis<12)
                            octaveLines[numOfFinalLine].lineIDInScaleLineVec = octaveLines[closeLineID].lineIDInScaleLineVec;
                        else{
                            octaveLines[numOfFinalLine].lineIDInScaleLineVec = lineIDInScaleLineVec;
                            lineIDInScaleLineVec++;
                        }

                        octaveLines[numOfFinalLine].octaveCount = octaveCount;
                        octaveLines[numOfFinalLine].lineIDInOctave = lineCurId;
                        octaveLines[numOfFinalLine].lineLength = length;
                        numOfFinalLine++;
                    }
                }
    }

    /* Reorganize the detected lines into keyLines */
    keyLines.clear();
    keyLines.resize(lineIDInScaleLineVec);
    unsigned int tempID;
    float s1,e1,s2,e2;
    bool shouldChange;
    OctaveSingleLine singleLine;
    for(unsigned int  lineID = 0;lineID < numOfFinalLine; lineID++)
    {
        lineIDInOctave = octaveLines[lineID].lineIDInOctave;
        octaveID  = octaveLines[lineID].octaveCount;

        cv::Vec3i tempParams;
        getLineParameters(extractedLines[octaveID][lineIDInOctave], tempParams);
        direction = getLineDirection(tempParams);
        // direction = edLineVec_[octaveID]->lineDirection_[lineIDInOctave];
        singleLine.octaveCount = octaveID;
        singleLine.direction = direction;
        singleLine.lineLength = octaveLines[lineID].lineLength;
        //singleLine.salience = singleLine.lineLength/max(images_sizes[octaveID].width,
        //                                               images_sizes[octaveID].height);
        //singleLine.salience  = edLineVec_[octaveID]->lineSalience_[lineIDInOctave];
        //singleLine.numOfPixels = edLineVec_[octaveID]->lines_.sId[lineIDInOctave+1]-
                                 //edLineVec_[octaveID]->lines_.sId[lineIDInOctave];

        //decide the start point and end point
        shouldChange = false;
        s1 = (extractedLines[octaveID][lineIDInOctave])[0];//sx
        s2 = (extractedLines[octaveID][lineIDInOctave])[1];//sy
        e1 = (extractedLines[octaveID][lineIDInOctave])[2];//ex
        e2 = (extractedLines[octaveID][lineIDInOctave])[3];//ey
        dx = e1 - s1;//ex-sx
        dy = e2 - s2;//ey-sy

        if(direction>=-0.75*M_PI&&direction<-0.25*M_PI)
        {
            if(dy>0)
                shouldChange = true;
        }

        if(direction>=-0.25*M_PI&&direction<0.25*M_PI)
            if(dx<0){shouldChange = true;}

        if(direction>=0.25*M_PI&&direction<0.75*M_PI)
            if(dy<0){shouldChange = true;}

        if((direction>=0.75*M_PI&&direction<M_PI)||(direction>=-M_PI&&direction<-0.75*M_PI)){
            if(dx>0)
                shouldChange = true;
        }

        tempValue = scale[octaveID];

        if(shouldChange){
                singleLine.sPointInOctaveX = e1;
                singleLine.sPointInOctaveY = e2;
                singleLine.ePointInOctaveX = s1;
                singleLine.ePointInOctaveY = s2;
                singleLine.startPointX = tempValue * e1;
                singleLine.startPointY = tempValue * e2;
                singleLine.endPointX   = tempValue * s1;
                singleLine.endPointY   = tempValue * s2;
            }

        else{
                singleLine.sPointInOctaveX = s1;
                singleLine.sPointInOctaveY = s2;
                singleLine.ePointInOctaveX = e1;
                singleLine.ePointInOctaveY = e2;
                singleLine.startPointX = tempValue * s1;
                singleLine.startPointY = tempValue * s2;
                singleLine.endPointX   = tempValue * e1;
                singleLine.endPointY   = tempValue * e2;
            }

        tempID = octaveLines[lineID].lineIDInScaleLineVec;
        keyLines[tempID].push_back(singleLine);
    }

    delete [] scale;

    return 1;

}

/* compute LBD descriptors */
int BinaryDescriptor::ComputeLBD_(ScaleLines &keyLines)
{
    //the default length of the band is the line length.
    short numOfFinalLine = keyLines.size();
    float *dL = new float[2];//line direction cos(dir), sin(dir)
    float *dO = new float[2];//the clockwise orthogonal vector of line direction.
    short heightOfLSP = params.widthOfBand_*params.numOfBand_;//the height of line support region;
    short descriptorSize = params.numOfBand_ * 8;//each band, we compute the m( pgdL, ngdL,  pgdO, ngdO) and std( pgdL, ngdL,  pgdO, ngdO);
    float pgdLRowSum;//the summation of {g_dL |g_dL>0 } for each row of the region;
    float ngdLRowSum;//the summation of {g_dL |g_dL<0 } for each row of the region;
    float pgdL2RowSum;//the summation of {g_dL^2 |g_dL>0 } for each row of the region;
    float ngdL2RowSum;//the summation of {g_dL^2 |g_dL<0 } for each row of the region;
    float pgdORowSum;//the summation of {g_dO |g_dO>0 } for each row of the region;
    float ngdORowSum;//the summation of {g_dO |g_dO<0 } for each row of the region;
    float pgdO2RowSum;//the summation of {g_dO^2 |g_dO>0 } for each row of the region;
    float ngdO2RowSum;//the summation of {g_dO^2 |g_dO<0 } for each row of the region;

    float *pgdLBandSum  = new float[params.numOfBand_];//the summation of {g_dL |g_dL>0 } for each band of the region;
    float *ngdLBandSum  = new float[params.numOfBand_];//the summation of {g_dL |g_dL<0 } for each band of the region;
    float *pgdL2BandSum = new float[params.numOfBand_];//the summation of {g_dL^2 |g_dL>0 } for each band of the region;
    float *ngdL2BandSum = new float[params.numOfBand_];//the summation of {g_dL^2 |g_dL<0 } for each band of the region;
    float *pgdOBandSum  = new float[params.numOfBand_];//the summation of {g_dO |g_dO>0 } for each band of the region;
    float *ngdOBandSum  = new float[params.numOfBand_];//the summation of {g_dO |g_dO<0 } for each band of the region;
    float *pgdO2BandSum = new float[params.numOfBand_];//the summation of {g_dO^2 |g_dO>0 } for each band of the region;
    float *ngdO2BandSum = new float[params.numOfBand_];//the summation of {g_dO^2 |g_dO<0 } for each band of the region;

    short numOfBitsBand = params.numOfBand_*sizeof(float);
    short lengthOfLSP; //the length of line support region, varies with lines
    short halfHeight = (heightOfLSP-1)/2;
    short halfWidth;
    short bandID;
    float coefInGaussion;
    float lineMiddlePointX, lineMiddlePointY;
    float sCorX, sCorY,sCorX0, sCorY0;
    short tempCor, xCor, yCor;//pixel coordinates in image plane
    short dx, dy;
    float gDL;//store the gradient projection of pixels in support region along dL vector
    float gDO;//store the gradient projection of pixels in support region along dO vector
    short imageWidth, imageHeight, realWidth;
    short *pdxImg, *pdyImg;
    float *desVec;

    short sameLineSize;
    short octaveCount;
    OctaveSingleLine *pSingleLine;
    /* loop over list of LineVec */
    for(short lineIDInScaleVec = 0; lineIDInScaleVec<numOfFinalLine; lineIDInScaleVec++){
        sameLineSize = keyLines[lineIDInScaleVec].size();
        /* loop over current LineVec's lines */
        for(short lineIDInSameLine = 0; lineIDInSameLine<sameLineSize; lineIDInSameLine++){
            /* get a line in current LineVec and its original ID in its octave */
            pSingleLine = &(keyLines[lineIDInScaleVec][lineIDInSameLine]);
            octaveCount = pSingleLine->octaveCount;

            /* retrieve associated dxImg and dyImg
            pdxImg = edLineVec_[octaveCount]->dxImg_.ptr<short>();
            pdyImg = edLineVec_[octaveCount]->dyImg_.ptr<short>(); */
            pdxImg = dxImg_vector[octaveCount].ptr<short>();
            pdyImg = dyImg_vector[octaveCount].ptr<short>();

            /* get image size to work on from real one
            realWidth = edLineVec_[octaveCount]->imageWidth;
            imageWidth  = realWidth -1;
            imageHeight = edLineVec_[octaveCount]->imageHeight-1; */
            realWidth = images_sizes[octaveCount].width;
            imageWidth = realWidth - 1;
            imageHeight = images_sizes[octaveCount].height - 1;


            /* initialize memory areas */
            memset(pgdLBandSum,  0, numOfBitsBand);
            memset(ngdLBandSum, 0, numOfBitsBand);
            memset(pgdL2BandSum,  0, numOfBitsBand);
            memset(ngdL2BandSum, 0, numOfBitsBand);
            memset(pgdOBandSum,  0, numOfBitsBand);
            memset(ngdOBandSum, 0, numOfBitsBand);
            memset(pgdO2BandSum,  0, numOfBitsBand);
            memset(ngdO2BandSum, 0, numOfBitsBand);

            /* get length of line and its half */
            lengthOfLSP = keyLines[lineIDInScaleVec][lineIDInSameLine].numOfPixels;
            halfWidth   = (lengthOfLSP-1)/2;

            /* get middlepoint of line */
            lineMiddlePointX = 0.5 * (pSingleLine->sPointInOctaveX +  pSingleLine->ePointInOctaveX);
            lineMiddlePointY = 0.5 * (pSingleLine->sPointInOctaveY +  pSingleLine->ePointInOctaveY);

            /*1.rotate the local coordinate system to the line direction (direction is the angle
                between positive line direction and positive X axis)
             *2.compute the gradient projection of pixels in line support region*/

            /* get the vector representing original image reference system after rotation to aligh with
               line's direction */
            dL[0] = cos(pSingleLine->direction);
            dL[1] = sin(pSingleLine->direction);

            /* set the clockwise orthogonal vector of line direction */
            dO[0] = -dL[1];
            dO[1] = dL[0];

            /* get rotated reference frame */
            sCorX0= -dL[0]*halfWidth + dL[1]*halfHeight + lineMiddlePointX;//hID =0; wID = 0;
            sCorY0= -dL[1]*halfWidth - dL[0]*halfHeight + lineMiddlePointY;


            /* BIAS::Matrix<float> gDLMat(heightOfLSP,lengthOfLSP) */
            for(short hID = 0; hID <heightOfLSP; hID++){
                /*initialization */
                sCorX = sCorX0;
                sCorY = sCorY0;

                pgdLRowSum = 0;
                ngdLRowSum = 0;
                pgdORowSum = 0;
                ngdORowSum = 0;

                for(short wID = 0; wID <lengthOfLSP; wID++){
                    tempCor = round(sCorX);
                    xCor = (tempCor<0)?0:(tempCor>imageWidth)?imageWidth:tempCor;
                    tempCor = round(sCorY);
                    yCor = (tempCor<0)?0:(tempCor>imageHeight)?imageHeight:tempCor;

                    /* To achieve rotation invariance, each simple gradient is rotated aligned with
                     * the line direction and clockwise orthogonal direction.*/
                    dx = pdxImg[yCor*realWidth+xCor];
                    dy = pdyImg[yCor*realWidth+xCor];
                    gDL = dx * dL[0] + dy * dL[1];
                    gDO = dx * dO[0] + dy * dO[1];
                    if(gDL>0){
                        pgdLRowSum  += gDL;
                    }else{
                        ngdLRowSum  -= gDL;
                    }
                    if(gDO>0){
                        pgdORowSum  += gDO;
                    }else{
                        ngdORowSum  -= gDO;
                    }
                    sCorX +=dL[0];
                    sCorY +=dL[1];
                    /* gDLMat[hID][wID] = gDL; */
                }
                sCorX0 -=dL[1];
                sCorY0 +=dL[0];
                coefInGaussion = gaussCoefG_[hID];
                pgdLRowSum = coefInGaussion * pgdLRowSum;
                ngdLRowSum = coefInGaussion * ngdLRowSum;
                pgdL2RowSum = pgdLRowSum * pgdLRowSum;
                ngdL2RowSum = ngdLRowSum * ngdLRowSum;
                pgdORowSum = coefInGaussion * pgdORowSum;
                ngdORowSum = coefInGaussion * ngdORowSum;
                pgdO2RowSum = pgdORowSum * pgdORowSum;
                ngdO2RowSum = ngdORowSum * ngdORowSum;

                /* compute {g_dL |g_dL>0 }, {g_dL |g_dL<0 },
                {g_dO |g_dO>0 }, {g_dO |g_dO<0 } of each band in the line support region
                first, current row belong to current band */
                bandID = hID/params.widthOfBand_;
                coefInGaussion = gaussCoefL_[hID%params.widthOfBand_+params.widthOfBand_];
                pgdLBandSum[bandID] +=  coefInGaussion * pgdLRowSum;
                ngdLBandSum[bandID] +=  coefInGaussion * ngdLRowSum;
                pgdL2BandSum[bandID] +=  coefInGaussion * coefInGaussion * pgdL2RowSum;
                ngdL2BandSum[bandID] +=  coefInGaussion * coefInGaussion * ngdL2RowSum;
                pgdOBandSum[bandID] +=  coefInGaussion * pgdORowSum;
                ngdOBandSum[bandID] +=  coefInGaussion * ngdORowSum;
                pgdO2BandSum[bandID] +=  coefInGaussion * coefInGaussion * pgdO2RowSum;
                ngdO2BandSum[bandID] +=  coefInGaussion * coefInGaussion * ngdO2RowSum;

                /* In order to reduce boundary effect along the line gradient direction,
                 * a row's gradient will contribute not only to its current band, but also
                 * to its nearest upper and down band with gaussCoefL_.*/
                bandID--;
                if(bandID>=0){/* the band above the current band */
                    coefInGaussion = gaussCoefL_[hID%params.widthOfBand_ + 2*params.widthOfBand_];
                    pgdLBandSum[bandID] +=  coefInGaussion * pgdLRowSum;
                    ngdLBandSum[bandID] +=  coefInGaussion * ngdLRowSum;
                    pgdL2BandSum[bandID] +=  coefInGaussion * coefInGaussion * pgdL2RowSum;
                    ngdL2BandSum[bandID] +=  coefInGaussion * coefInGaussion * ngdL2RowSum;
                    pgdOBandSum[bandID] +=  coefInGaussion * pgdORowSum;
                    ngdOBandSum[bandID] +=  coefInGaussion * ngdORowSum;
                    pgdO2BandSum[bandID] +=  coefInGaussion * coefInGaussion * pgdO2RowSum;
                    ngdO2BandSum[bandID] +=  coefInGaussion * coefInGaussion * ngdO2RowSum;
                }
                bandID = bandID+2;
                if(bandID<params.numOfBand_){/*the band below the current band */
                    coefInGaussion = gaussCoefL_[hID%params.widthOfBand_];
                    pgdLBandSum[bandID] +=  coefInGaussion * pgdLRowSum;
                    ngdLBandSum[bandID] +=  coefInGaussion * ngdLRowSum;
                    pgdL2BandSum[bandID] +=  coefInGaussion * coefInGaussion * pgdL2RowSum;
                    ngdL2BandSum[bandID] +=  coefInGaussion * coefInGaussion * ngdL2RowSum;
                    pgdOBandSum[bandID] +=  coefInGaussion * pgdORowSum;
                    ngdOBandSum[bandID] +=  coefInGaussion * ngdORowSum;
                    pgdO2BandSum[bandID] +=  coefInGaussion * coefInGaussion * pgdO2RowSum;
                    ngdO2BandSum[bandID] +=  coefInGaussion * coefInGaussion * ngdO2RowSum;
                }
            }
            /* gDLMat.Save("gDLMat.txt");
            return 0; */

            /* construct line descriptor */
            pSingleLine->descriptor.resize(descriptorSize);
            desVec = pSingleLine->descriptor.data();

            short desID;

            /*Note that the first and last bands only have (lengthOfLSP * widthOfBand_ * 2.0) pixels
             * which are counted. */
            float invN2 = 1.0/(params.widthOfBand_ * 2.0);
            float invN3 = 1.0/(params.widthOfBand_ * 3.0);
            float invN, temp;
            for(bandID = 0; bandID<params.numOfBand_; bandID++){
                if(bandID==0||bandID==params.numOfBand_-1){	invN = invN2;
                }else{ invN = invN3;}
                desID = bandID * 8;
                temp = pgdLBandSum[bandID] * invN;
                desVec[desID]   = temp;/* mean value of pgdL; */
                desVec[desID+4] = sqrt(pgdL2BandSum[bandID] * invN - temp*temp);//std value of pgdL;
                temp = ngdLBandSum[bandID] * invN;
                desVec[desID+1] = temp;//mean value of ngdL;
                desVec[desID+5] = sqrt(ngdL2BandSum[bandID] * invN - temp*temp);//std value of ngdL;

                temp = pgdOBandSum[bandID] * invN;
                desVec[desID+2] = temp;//mean value of pgdO;
                desVec[desID+6] = sqrt(pgdO2BandSum[bandID] * invN - temp*temp);//std value of pgdO;
                temp = ngdOBandSum[bandID] * invN;
                desVec[desID+3] = temp;//mean value of ngdO;
                desVec[desID+7] = sqrt(ngdO2BandSum[bandID] * invN - temp*temp);//std value of ngdO;
            }

            // normalize;
            float tempM, tempS;
            tempM = 0;
            tempS = 0;
            desVec = pSingleLine->descriptor.data();

            int base = 0;
            for(short i=0; i<params.numOfBand_*8; ++base, i=base*8){
                tempM += *(desVec+i) * *(desVec+i);//desVec[8*i+0] * desVec[8*i+0];
                tempM += *(desVec+i+1) * *(desVec+i+1);//desVec[8*i+1] * desVec[8*i+1];
                tempM += *(desVec+i+2) * *(desVec+i+2);//desVec[8*i+2] * desVec[8*i+2];
                tempM += *(desVec+i+3) * *(desVec+i+3);//desVec[8*i+3] * desVec[8*i+3];
                tempS += *(desVec+i+4) * *(desVec+i+4);//desVec[8*i+4] * desVec[8*i+4];
                tempS += *(desVec+i+5) * *(desVec+i+5);//desVec[8*i+5] * desVec[8*i+5];
                tempS += *(desVec+i+6) * *(desVec+i+6);//desVec[8*i+6] * desVec[8*i+6];
                tempS += *(desVec+i+7) * *(desVec+i+7);//desVec[8*i+7] * desVec[8*i+7];
            }

            tempM = 1/sqrt(tempM);
            tempS = 1/sqrt(tempS);
            desVec = pSingleLine->descriptor.data();
            base = 0;
            for(short i=0; i<params.numOfBand_*8; ++base, i=base*8){
                *(desVec+i) = *(desVec+i) * tempM;//desVec[8*i] =  desVec[8*i] * tempM;
                *(desVec+1+i) = *(desVec+1+i) * tempM;//desVec[8*i+1] =  desVec[8*i+1] * tempM;
                *(desVec+2+i) = *(desVec+2+i) * tempM;//desVec[8*i+2] =  desVec[8*i+2] * tempM;
                *(desVec+3+i) = *(desVec+3+i) * tempM;//desVec[8*i+3] =  desVec[8*i+3] * tempM;
                *(desVec+4+i) = *(desVec+4+i) * tempS;//desVec[8*i+4] =  desVec[8*i+4] * tempS;
                *(desVec+5+i) = *(desVec+5+i) * tempS;//desVec[8*i+5] =  desVec[8*i+5] * tempS;
                *(desVec+6+i) = *(desVec+6+i) * tempS;//desVec[8*i+6] =  desVec[8*i+6] * tempS;
                *(desVec+7+i) = *(desVec+7+i) * tempS;//desVec[8*i+7] =  desVec[8*i+7] * tempS;
            }

            /* In order to reduce the influence of non-linear illumination,
             * a threshold is used to limit the value of element in the unit feature
             * vector no larger than this threshold. In Z.Wang's work, a value of 0.4 is found
             * empirically to be a proper threshold.*/
            desVec = pSingleLine->descriptor.data();
            for(short i=0; i<descriptorSize; i++ ){
                if(desVec[i]>0.4){
                    desVec[i]=0.4;
                }
            }

            //re-normalize desVec;
            temp = 0;
            for(short i=0; i<descriptorSize; i++){
                temp += desVec[i] * desVec[i];
            }

            temp = 1/sqrt(temp);
            for(short i=0; i<descriptorSize; i++){
                desVec[i] =  desVec[i] * temp;
            }
        }/* end for(short lineIDInSameLine = 0; lineIDInSameLine<sameLineSize;
            lineIDInSameLine++) */

    }/* end for(short lineIDInScaleVec = 0;
        lineIDInScaleVec<numOfFinalLine; lineIDInScaleVec++) */

    delete [] dL;
    delete [] dO;
    delete [] pgdLBandSum;
    delete [] ngdLBandSum;
    delete [] pgdL2BandSum;
    delete [] ngdL2BandSum;
    delete [] pgdOBandSum;
    delete [] ngdOBandSum;
    delete [] pgdO2BandSum;
    delete [] ngdO2BandSum;

    return 1;
}


